/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type ComputeSystem struct {
	Owner string `json:"Owner,omitempty"`

	SchemaVersion *Version `json:"SchemaVersion,omitempty"`

	HostingSystemId string `json:"HostingSystemId,omitempty"`

	HostedSystem interface{} `json:"HostedSystem,omitempty"`

	Container *Container `json:"Container,omitempty"`

	VirtualMachine *VirtualMachine `json:"VirtualMachine,omitempty"`

	ShouldTerminateOnLastHandleClosed bool `json:"ShouldTerminateOnLastHandleClosed,omitempty"`
}
